import sys
import random
import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import warnings
import joblib
import os
from multiprocessing import Pool
from itertools import repeat
import pickle
from sklearn.decomposition import LatentDirichletAllocation
from scipy import stats as st
import scanpy.external as sce
import networkx as nx
import math

from Topyfic.train import *
from Topyfic.topic import *
from Topyfic.topModel import *
from Topyfic.analysis import *

warnings.filterwarnings("ignore")


def train_model(name,
                data,
                k,
                n_runs=100,
                random_state_range=None,
                n_thread=5,
                save_path=""):
    """
    Training model and save it

    :param name: name of the Train class
    :type name: str
    :param k: number of topics to learn one LDA model using sklearn package (default: 50)
    :type k: int
    :param n_runs: number of run to define rLDA model (default: 100)
    :type n_runs: int
    :param random_state_range: list of random state, we used to run LDA models (default: range(n_runs))
    :type random_state_range: list of int
    :param data: data embedded in anndata format use to train LDA model
    :type data: anndata
    :param n_thread: number of threads you used to learn LDA models (default=5)
    :type n_thread: int
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str

    """
    train = Train(name=name,
                  k=k,
                  n_runs=n_runs,
                  random_state_range=random_state_range)
    train.run_LDA_models(data, n_thread=n_thread)
    train.save_train(save_path=save_path)


def make_topModel(trains,
                  data,
                  n_top_genes=50,
                  resolution=1,
                  file_format="pdf",
                  save_path=""):
    """
    Creating topModel base on train data and save it along with clustering information

    :param trains: list of train class
    :type trains: list of Train
    :param data: expression data embedded in anndata format along with cells and genes/region information
    :type data: anndata
    :param n_top_genes: Number of highly-variable genes to keep (default: 50)
    :type n_top_genes: int
    :param resolution: A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. (default: 1)
    :type resolution: int
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str

    """
    top_model, clustering = calculate_leiden_clustering(trains=trains,
                                                        data=data,
                                                        n_top_genes=n_top_genes,
                                                        resolution=resolution,
                                                        file_format=file_format)
    clustering.to_csv(f"{save_path}/clustering.csv")
    top_model.save_topModel(save_path=save_path)


def make_analysis_class(top_model,
                        data,
                        colors_topics=None,
                        save_path=""):
    """
    Creating Analysis object

    :param top_model: top model that used for analysing topics, gene weights compositions and calculate cell participation
    :type top_model: TopModel
    :param data: processed expression data along with cells and genes/region information
    :type data: anndata
    :param colors_topics: dataframe that mapped colored to topics
    :type colors_topics: pandas dataframe
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str
    """
    analysis_top_model = Analysis(Top_model=top_model, colors_topics=colors_topics)
    analysis_top_model.calculate_cell_participation(data=data)
    analysis_top_model.save_analysis(save_path=save_path)


def subset_data(data, keep, loc='var'):
    """
    Subsetting data

    :param data: data we want to subset
    :type data: anndata
    :param keep: values in the obs/var_names
    :type keep: list
    :param loc: subsetting in which direction (default: 'var')

    :return: data we want to keep
    :rtype: anndata
    """
    if loc == 'var':
        data = data[:, keep]
    elif loc == 'obs':
        data = data[keep, :]
    else:
        sys.exit("loc is not correct! it should be 'obs' or 'var'")

    return data


def calculate_leiden_clustering(trains,
                                data,
                                n_top_genes=50,
                                resolution=1,
                                min_cell_participation=None,
                                file_format="pdf"):
    """
    Do leiden clustering w/o harmony base on number of assays you have and then remove low participation topics

    :param trains: list of train class
    :type trains: list of Train
    :param data: gene-count data with cells and genes information
    :type data: anndata
    :param n_top_genes: Number of highly-variable genes to keep (default: 50)
    :type n_top_genes: int
    :param resolution: A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. (default: 1)
    :type resolution: int
    :param min_cell_participation: minimum cell participation across for each topics to keep them, when is None, it will keep topics with cell participation more than number of cell / 10000.
    :type min_cell_participation: float
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str

    :return: final TopModel instance after clustering and trimming, dataframe containing which run goes to which topic
    :rtype: TopModel, pandas dataframe
    """
    all_batches = None
    all_components = None
    all_exp_dirichlet_component = None
    all_others = None

    if min_cell_participation is None:
        min_cell_participation = data.to_df().shape[0] / 10000

    for train in trains:
        components, exp_dirichlet_component, others = train.make_LDA_models_attributes()
        components.index = f"{train.name}_" + components.index
        exp_dirichlet_component.index = f"{train.name}_" + exp_dirichlet_component.index
        others.index = f"{train.name}_" + others.index
        batch = [train.name] * components.shape[0]
        if all_components is None:
            all_batches = batch
            all_components = components
            all_exp_dirichlet_component = exp_dirichlet_component
            all_others = others
        else:
            all_batches = batch + all_batches
            all_components = pd.concat([components, all_components], axis=0)
            all_exp_dirichlet_component = pd.concat([exp_dirichlet_component, all_exp_dirichlet_component], axis=0)
            all_others = pd.concat([others, all_others], axis=0)

    if len(trains) == 1:
        adata = anndata.AnnData(all_components)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes)
        sc.pp.neighbors(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=resolution)
        adata.obs["topics"] = adata.obs["leiden"].astype(int) + 1
        adata.obs["topics"] = "Topic_" + adata.obs["topics"].astype(str)
        sc.pl.umap(adata, color=["topics"],
                   title=[f"Topic space UMAP leiden clusters (k={trains[0].k})"],
                   save=f"_leiden_clustering.{file_format}")
        sc.pl.umap(adata, color=['topics'],
                   legend_loc='on data',
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_v2.{file_format}')

    else:
        adata = anndata.AnnData(all_components)
        adata.obs['assays'] = all_batches
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes)
        sc.pp.neighbors(adata)
        sce.pp.harmony_integrate(adata, 'assays')
        adata.obsm['X_pca'] = adata.obsm['X_pca_harmony']
        sc.pp.neighbors(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=resolution)
        adata.obs["topics"] = adata.obs["leiden"].astype(int) + 1
        adata.obs["topics"] = "Topic_" + adata.obs["topics"].astype(str)
        sc.pl.umap(adata, color=['topics'],
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_harmony.{file_format}')
        sc.pl.umap(adata, color=['topics'],
                   legend_loc='on data',
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_harmony_v2.{file_format}')
        sc.pl.umap(adata, color=['assays'],
                   title=['Topic space UMAP leiden clusters'],
                   save=f'_technology_harmony.{file_format}')

    clustering = adata.obs.copy(deep=True)
    clustering["leiden"] = clustering["leiden"].astype(int)
    n_rtopics = len(np.unique(clustering["leiden"]))

    rlda = initialize_rLDA_model(all_components,
                                 all_exp_dirichlet_component,
                                 all_others,
                                 clusters=clustering)
    lda_output = rlda.transform(data.X)
    cell_participation = pd.DataFrame(np.round(lda_output, 2),
                                      columns=[f"Topic{i + 1}" for i in range(n_rtopics)],
                                      index=data.obs.index)

    keep = cell_participation.sum() > min_cell_participation
    print(
        f"{keep.sum()} topics out of {keep.shape[0]} topics have participation more than {min_cell_participation}")
    tmp = pd.DataFrame(keep)
    tmp["cluster"] = range(tmp.shape[0])
    tmp = tmp.to_dict()
    tmp = tmp[0]
    clustering["keep"] = clustering["leiden"].astype(int) + 1
    clustering["keep"] = "Topic" + clustering["keep"].astype(str)
    clustering["keep"] = clustering["keep"].replace(tmp)
    n_rtopics = keep.sum()
    rlda, gene_weights_T = filter_LDA_model(rlda, keep)

    gene_weights = gene_weights_T.T
    gene_weights.index = data.var.index

    if len(trains) == 1:
        top_model = TopModel(name=trains[0].name,
                             N=n_rtopics,
                             gene_weights=gene_weights,
                             rlda=rlda)

    else:
        name = np.unique(all_batches).tolist()
        name = '_'.join(name)

        top_model = TopModel(name=name,
                             N=n_rtopics,
                             gene_weights=gene_weights,
                             rlda=rlda)

    return top_model, clustering


def plot_cluster_contribution(clustering,
                              feature,
                              show_all=False,
                              portion=True,
                              save=True,
                              show=True,
                              file_format="pdf",
                              file_name='cluster_contribution'):
    """
    barplot shows number of topics contribute to each cluster

    :param clustering: dataframe that map each single LDA run to each cluster
    :type clustering: pandas dataframe
    :param feature: name of the feature you want to see the cluster contribution (should be one of the columns name of clustering df)
    :type feature: str
    :param show_all: Indicate if you want to show all clusters or only the ones that pass threshold (default: False)
    :type show_all: bool
    :param portion: Indicate if you want to normalized the bar to show percentage instead of actual value (default: True)
    :type portion: bool
    :param save: indicate if you want to save the plot or not (default: True)
    :type save: bool
    :param show: indicate if you want to show the plot or not (default: True)
    :type show: bool
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param file_name: name and path of the plot use for save (default: cluster_contribution)
    :type file_name: str
    """
    if feature not in clustering.columns:
        sys.exit(f"{feature} is not valid! should be a columns names of clustering")

    if not show_all:
        clustering = clustering[clustering["keep"]]

    options = np.unique(clustering[feature]).tolist()
    res = pd.DataFrame(columns=options,
                       index=range(len(clustering['leiden'].unique())))

    for i in range(res.shape[0]):
        for opt in options:
            tmp = clustering[np.logical_and(clustering['leiden'] == i,
                                            clustering[feature] == opt)]
            res.loc[i, opt] = tmp.shape[0]
    if not portion:
        plot = res.plot.bar(stacked=True,
                            xlabel='leiden',
                            ylabel='number of topics',
                            figsize=(max(res.shape[0] / 2, 5), 7))
    else:
        sum_feature = res.sum(axis=1)
        for i in range(res.shape[0]):
            for opt in options:
                res.loc[i, opt] = res.loc[i, opt] / sum_feature[i] * 100
        res.index = res.index + 1
        res.index = "Topics_" + res.index.astype(str)
        plot = res.plot.bar(stacked=True,
                            xlabel='leiden',
                            ylabel='percentage(%) of topics',
                            figsize=(max(res.shape[0] / 2, 5), 7))
    fig = plot.get_figure()
    if save:
        fig.savefig(f"{file_name}.{file_format}", bbox_inches='tight')
    if show:
        plt.show()
    else:
        plt.close()


def initialize_rLDA_model(all_components, all_exp_dirichlet_component, all_others, clusters):
    """
    Initialize reproducible LDA model by calculating all necessary attributes using clustering.

    :param all_components: Variational parameters for topic gene distribution from all single LDA models
    :type all_components: pandas dataframe
    :param all_exp_dirichlet_component: Exponential value of expectation of log topic gene distribution from all single LDA models
    :type all_exp_dirichlet_component: pandas dataframe
    :param all_others: dataframe contains remaining necessary attributes including: n_batch_iter: Number of iterations of the EM step. n_features_in: Number of features seen during fit. n_iter: Number of passes over the dataset. bound: Final perplexity score on training set. doc_topic_prior: Prior of document topic distribution theta. If the value is None, it is 1 / n_components. topic_word_prior: Prior of topic word distribution beta. If the value is None, it is 1 / n_components.
    :type all_others: pandas dataframe
    :param clusters: dataframe that mapped each LDA run to each clusters
    :type clusters: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :rtype: sklearn.decomposition.LatentDirichletAllocation
    """
    n_rtopics = len(np.unique(clusters[f"leiden"]))

    components = np.zeros((n_rtopics, all_components.shape[1]), dtype=float)
    exp_dirichlet_component = np.zeros((n_rtopics, all_exp_dirichlet_component.shape[1]), dtype=float)

    topic = 0
    for cluster in np.unique(clusters[f"leiden"]):
        tmp = clusters[clusters["leiden"] == cluster]

        tmp_components = all_components.loc[tmp.index, :]
        tmp_components = tmp_components.mean(axis=0)
        components[topic, :] = tmp_components.values

        tmp_exp_dirichlet_component = all_exp_dirichlet_component.loc[tmp.index, :]
        tmp_exp_dirichlet_component = tmp_exp_dirichlet_component.mean(axis=0)
        exp_dirichlet_component[topic, :] = tmp_exp_dirichlet_component.values

        topic = topic + 1

    df = pd.DataFrame(np.transpose(components),
                      columns=[f"Topic{i + 1}" for i in range(n_rtopics)],
                      index=all_components.columns)
    for topic in range(n_rtopics):
        df_sorted = df.sort_values([f"Topic{topic + 1}"], axis=0, ascending=False)[f"Topic{topic + 1}"]
        tmp = df_sorted.cumsum()
        tmp = tmp[tmp > 0.9 * df_sorted.sum()]
        df.loc[tmp.index, f"Topic{topic + 1}"] = 1 / n_rtopics

    components = df.T

    exp_dirichlet_component = pd.DataFrame(exp_dirichlet_component,
                                           index=[f"Topic{i + 1}" for i in range(n_rtopics)],
                                           columns=all_exp_dirichlet_component.columns)
    others = all_others.mean(axis=0)

    LDA = initialize_lda_model(components, exp_dirichlet_component, others)

    return LDA


def initialize_lda_model(components, exp_dirichlet_component, others):
    """
    Initialize LDA model by passing all necessary attributes

    :param components: Variational parameters for topic gene distribution
    :type components: pandas dataframe
    :param exp_dirichlet_component: Exponential value of expectation of log topic gene distribution
    :type exp_dirichlet_component: pandas dataframe
    :param others: dataframe contains remaining necessary attributes including: n_batch_iter: Number of iterations of the EM step. n_features_in: Number of features seen during fit. n_iter: Number of passes over the dataset. bound: Final perplexity score on training set. doc_topic_prior: Prior of document topic distribution theta. If the value is None, it is 1 / n_components. topic_word_prior: Prior of topic word distribution beta. If the value is None, it is 1 / n_components.
    :type others: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :rtype: sklearn.decomposition.LatentDirichletAllocation
    """
    n_topics = components.shape[0]

    LDA = LatentDirichletAllocation(n_components=n_topics)

    LDA.components_ = components.values
    LDA.exp_dirichlet_component_ = exp_dirichlet_component.values
    LDA.n_batch_iter_ = int(others['n_batch_iter'])
    LDA.n_features_in_ = int(others['n_features_in'])
    LDA.n_iter_ = int(others['n_iter'])
    LDA.bound_ = others['bound']
    LDA.doc_topic_prior_ = others['doc_topic_prior']
    LDA.topic_word_prior_ = others['topic_word_prior']

    return LDA


def filter_LDA_model(main_lda, keep):
    """
    filter LDA based on the topics we want to keep

    :param main_lda: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :type main_lda: sklearn.decomposition.LatentDirichletAllocation
    :param keep: dataframe that define which topics we want to keep
    :type keep: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm, weights of genes in each topics (indexes are topics and columns are genes)
    :rtype: sklearn.decomposition.LatentDirichletAllocation, pandas dataframe
    """
    n_topics = keep.sum()
    lda = LatentDirichletAllocation(n_components=n_topics)

    components = main_lda.components_[[i for i in range(keep.shape[0]) if keep[i]], :]

    df = pd.DataFrame(np.transpose(components),
                      columns=[f'Topic{i + 1}' for i in range(n_topics)])
    for topic in range(n_topics):
        df_sorted = df.sort_values([f'Topic{topic + 1}'], axis=0, ascending=False)[f'Topic{topic + 1}']
        tmp = df_sorted.cumsum()
        tmp = tmp[tmp > 0.9 * df_sorted.sum()]
        df.loc[tmp.index, f'Topic{topic + 1}'] = 1 / n_topics

    components = df.T

    lda.components_ = components.values
    lda.exp_dirichlet_component_ = main_lda.exp_dirichlet_component_[[i for i in range(keep.shape[0]) if keep[i]], :]
    lda.n_batch_iter_ = main_lda.n_batch_iter_
    lda.n_features_in_ = main_lda.n_features_in_
    lda.n_iter_ = main_lda.n_iter_
    lda.bound_ = main_lda.bound_
    lda.doc_topic_prior_ = main_lda.doc_topic_prior_
    lda.topic_word_prior_ = main_lda.topic_word_prior_

    return lda, components


def read_train(file):
    """
    reading train pickle file

    :param file: path of the pickle file
    :type file: str

    :return: train instance
    :rtype: Train class
    """
    if not os.path.isfile(file):
        raise ValueError('Train object not found at given path!')

    picklefile = open(file, 'rb')
    train = pickle.load(picklefile)

    print(f"Reading Train done!")
    return train


def read_topModel(file):
    """
    reading topModel pickle file

    :param file: path of the pickle file
    :type file: str

    :return: topModel instance
    :rtype: TopModel class
    """
    if not os.path.isfile(file):
        raise ValueError('TopModel object not found at given path!')

    picklefile = open(file, 'rb')
    topModel = pickle.load(picklefile)

    print(f"Reading TopModel done!")
    return topModel


def read_analysis(file):
    """
    reading analysis pickle file

    :param file: path of the pickle file
    :type file: str

    :return: analysis instance
    :rtype: Analysis class
    """
    if not os.path.isfile(file):
        raise ValueError('Analysis object not found at given path!')

    picklefile = open(file, 'rb')
    analysis = pickle.load(picklefile)

    print(f"Reading Analysis done!")
    return analysis


def compare_topModels(topModels,
                      output_type='graph',
                      threshold=0.8,
                      topModels_color=None,
                      topModels_label=None,
                      save=False,
                      plot_show=True,
                      figsize=None,
                      plot_format="pdf",
                      file_name="compare_topics"):
    """
    compare several topModels

    :param topModels: list of topModel class you want to compare to each other
    :type topModels: list of TopModel class
    :param output_type: indicate the type of output you want. graph: plot as a graph, heatmap: plot as a heatmap, table: table contains correlation
    :type output_type: str
    :param threshold: only apply when you choose circular which only show correlation above that
    :type threshold: float
    :param topModels_color: dictionary of colors mapping each topics to each color (default: blue)
    :type topModels_color: dict
    :param topModels_label: dictionary of label mapping each topics to each label
    :type topModels_label: dict
    :param save: indicate if you want to save the plot or not (default: True)
    :type save: bool
    :param plot_show: indicate if you want to show the plot or not (default: True)
    :type plot_show: bool
    :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
    :type figsize: tuple of int
    :param plot_format: indicate the format of plot (default: pdf)
    :type plot_format: str
    :param file_name: name and path of the plot use for save (default: piechart_topicAvgCell)
    :type file_name: str

    :return: table contains correlation between topics only when table is choose and save is False
    :rtype: pandas dataframe
    """
    if output_type not in ['graph', 'heatmap', 'table']:
        sys.exit("output_type is not valid! it should be one of 'graph', 'heatmap' or 'table'")

    all_gene_weights = None

    for topModel in topModels:
        gene_weights = topModel.get_gene_weights()
        if all_gene_weights is None:
            all_gene_weights = gene_weights
        else:
            all_gene_weights = pd.concat([gene_weights, all_gene_weights], axis=1)

    corrs = pd.DataFrame(index=all_gene_weights.columns,
                         columns=all_gene_weights.columns,
                         dtype='float64')

    for d1 in all_gene_weights.columns.tolist():
        for d2 in all_gene_weights.columns.tolist():
            if d1 == d2:
                corrs.at[d1, d2] = 1
                continue
            a = all_gene_weights[[d1, d2]]
            a.dropna(axis=0, how='all', inplace=True)
            a.fillna(0, inplace=True)
            corr = st.pearsonr(a[d1].tolist(), a[d2].tolist())
            corrs.at[d1, d2] = corr[0]

    if output_type == 'table':
        if save:
            corrs.to_csv(f"{file_name}.csv")
        else:
            return corrs

    if output_type == 'heatmap':
        sns.clustermap(corrs,
                       figsize=None)
        if save:
            plt.savefig(f"{file_name}.{plot_format}")
        if plot_show:
            plt.show()
        else:
            plt.close()

        return

    if output_type == 'graph':
        np.fill_diagonal(corrs.values, 0)
        corrs[corrs < threshold] = np.nan
        res = corrs.stack()
        res = pd.DataFrame(res)
        res.reset_index(inplace=True)
        res.columns = ['source', 'dest', 'weight']
        res['weight'] = res['weight'].astype(float).round(decimals=2)
        res['source_label'] = res['source']
        res['dest_label'] = res['dest']
        res['source_color'] = res['source']
        res['dest_color'] = res['dest']

        if topModels_label is not None:
            res['source_label'].replace(topModels_label, inplace=True)
            res['dest_label'].replace(topModels_label, inplace=True)
        if topModels_color is None:
            res['source_color'] = "blue"
            res['dest_color'] = "blue"
        else:
            res['source_color'].replace(topModels_color, inplace=True)
            res['dest_color'].replace(topModels_color, inplace=True)

        G = nx.Graph()
        for i in range(res.shape[0]):
            G.add_node(res.source_label[i], color=res.source_color[i])
            G.add_node(res.dest_label[i], color=res.dest_color[i])
            G.add_edge(res.source_label[i], res.dest_label[i], weight=res.weight[i])

        connected_components = sorted(nx.connected_components(G), key=len, reverse=True)

        if figsize is None:
            figsize = (len(connected_components) * 2, len(connected_components) * 2)

        nrows = math.ceil(math.sqrt(len(connected_components)))
        ncols = math.ceil(len(connected_components) / nrows)
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, facecolor='white')

        i = 0
        for connected_component in connected_components:
            g_connected_component = G.subgraph(connected_component)
            nodePos = nx.spring_layout(g_connected_component)

            edge_labels = nx.get_edge_attributes(g_connected_component, "weight")

            node_color = nx.get_node_attributes(g_connected_component, "color").values()
            weights = nx.get_edge_attributes(g_connected_component, 'weight').values()

            nx.draw_networkx(g_connected_component,
                             pos=nodePos,
                             width=list(weights),
                             with_labels=True,
                             node_color=list(node_color),
                             font_size=8,
                             node_size=500,
                             ax=axs[int(i / ncols), i % ncols])

            nx.draw_networkx_edge_labels(g_connected_component,
                                         nodePos,
                                         edge_labels=edge_labels,
                                         font_size=7,
                                         ax=axs[int(i / ncols), i % ncols])

            i += 1

        [axi.axis('off') for axi in axs.ravel()]
        plt.tight_layout()
        if save:
            plt.savefig(f"{file_name}.{plot_format}")
        if plot_show:
            plt.show()
        else:
            plt.close()

        return axs




import sys
import warnings
import joblib
import pandas as pd
import numpy as np

import gseapy as gp
from gseapy.plot import dotplot
from gseapy import gseaplot
from reactome2py import analysis

import matplotlib.pyplot as plt
import seaborn as sns

sns.set_context('paper')

warnings.filterwarnings('ignore')


class Topic:
    """
    A class saved topic along with other useful information

    :param topic_id: ID of topic which is unique
    :type topic_id: str
    :param topic_name: name of the topic (default: topic_id)
    :type topic_name: str
    :param topic_gene_weights: dataframe that contains weights of the topics for each genes
    :type topic_gene_weights: pandas dataframe
    :param gene_information: dataframe that contains information of genes i.e gene biotype
    :type gene_information: pandas dataframe
    :param topic_information: dataframe that contains information of genes i.e cell state / cell type
    :type topic_information: pandas dataframe
    """

    def __init__(self,
                 topic_id,
                 topic_name=None,
                 topic_gene_weights=None,
                 gene_information=None,
                 topic_information=None):

        self.id = topic_id

        if topic_name is not None:
            self.name = topic_name
        else:
            self.name = topic_id

        self.gene_weights = topic_gene_weights

        if gene_information is None:
            self.gene_information = pd.DataFrame(index=topic_gene_weights.index)
        else:
            self.gene_information = gene_information

        if topic_information is None:
            self.topic_information = pd.DataFrame(index=topic_gene_weights.columns)
        else:
            self.topic_information = topic_information

    def update_gene_information(self, gene_information):
        """
        update/add genes information for each topics

        :param gene_information: dataframe contains genes information we would like to add/update (the index should be same as an index of gene_information in class)
        :type gene_information: pandas dataframe
        """
        same_columns = self.gene_information.columns.intersection(gene_information.columns)
        self.gene_information.drop(same_columns, axis=1, inplace=True)
        self.gene_information = pd.concat([self.gene_information, gene_information], axis=1)

    def functional_enrichment_analysis(self,
                                       type,
                                       organism,
                                       sets=None,
                                       p_value=0.05,
                                       file_format="pdf",
                                       file_name="functional_enrichment_analysis"):
        """
        Doing functional enrichment analysis including GO, KEGG and REACTOME

        :param type: indicate the type of databases which it should be one of "GO", "REACTOME"
        :type type: str
        :param organism: name of the organ you want to do functional enrichment analysis
        :type organism: str
        :param sets: str, list, tuple of Enrichr Library name(s). (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats) only need to fill if the type is GO
        :type sets: str, list, tuple
        :param p_value: Defines the pValue threshold. (default: 0.05)
        :type p_value: float
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_composition)
        :type file_name: str
        """

        if type not in ["GO", "REACTOME"]:
            sys.exit("Type is not valid! it should be one of them GO, KEGG, REACTOME")

        if type == "GO" and sets is None:
            sets = ["GO_Biological_Process_2021"]
        elif type == "KEGG" and sets is None:
            sets = ["KEGG_2016"]

        gene_weights = self.gene_weights.copy(deep=True)
        gene_weights = gene_weights[gene_weights > gene_weights.min()]
        genes = gene_weights.dropna().index.tolist()
        if type in ["GO", "KEGG"]:
            enr = gp.enrichr(gene_list=genes,
                             gene_sets=sets,
                             organism=organism,
                             outdir=f"{file_name}",
                             cutoff=p_value)
            dotplot(enr.res2d,
                    title=f"Gene ontology in Topic {self.name}",
                    cmap='viridis_r',
                    cutoff=p_value,
                    ofname=f"{file_name}.{file_format}")
        else:
            numGeneModule = len(genes)
            genes = ",".join(genes)
            result = analysis.identifiers(ids=genes,
                                          species=organism,
                                          p_value=str(p_value))
            token = result['summary']['token']
            analysis.report(token,
                            path=f"{file_name}/",
                            file=f"{file_name}.{file_format}",
                            number='50',
                            species=organism)
            token_result = analysis.token(token,
                                          species=organism,
                                          p_value=str(p_value))

            print(
                f"{numGeneModule - token_result['identifiersNotFound']} out of {numGeneModule} identifiers in the sample were found in Reactome.")
            print(
                f"{token_result['resourceSummary'][0]['pathways']} pathways were hit by at least one of them, which {len(token_result['pathways'])} of them have p-value more than {p_value}.")
            print(f"Report was saved {file_name}!")
            print(f"For more information please visit https://reactome.org/PathwayBrowser/#/DTAB=AN&ANALYSIS={token}")

    def GSEA(self,
             gene_sets='GO_Biological_Process_2021',
             p_value=0.05,
             table=True,
             plot=True,
             file_format="pdf",
             file_name="GSEA",
             **kwargs):
        """
        Doing Gene Set Enrichment Analysis on based on the topic weights using GSEAPY package.

        :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats)
        :type gene_sets: str, list, tuple
        :param p_value: Defines the pValue threshold for plotting. (default: 0.05)
        :type p_value: float
        :param table: indicate if you want to save all GO terms that passed the threshold as a table (default: True)
        :type table: bool
        :param plot: indicate if you want to plot all GO terms that passed the threshold (default: True)
        :type plot: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_composition)
        :type file_name: str
        :param kwargs: Argument to pass to gseapy.prerank(). more info: https://gseapy.readthedocs.io/en/latest/run.html?highlight=gp.prerank#gseapy.prerank

        :return: dataframe contains these columns: Term: gene set name, ES: enrichment score, NES: normalized enrichment score, NOM p-val:  Nominal p-value (from the null distribution of the gene set, FDR q-val: FDR qvalue (adjusted False Discory Rate), FWER p-val: Family wise error rate p-values, Tag %: Percent of gene set before running enrichment peak (ES), Gene %: Percent of gene list before running enrichment peak (ES), Lead_genes: leading edge genes (gene hits before running enrichment peak)
        :rtype: pandas dataframe
        """

        gene_weights = self.gene_weights.copy(deep=True)
        gene_weights.index = gene_weights.index.str.upper()
        gene_weights.sort_values([self.id], ascending=False, inplace=True)
        gene_weights = gene_weights[gene_weights > gene_weights.min()]
        gene_weights.dropna(inplace=True)

        if gene_weights.shape[0] == 1:
            return

        pre_res = gp.prerank(rnk=gene_weights,
                             gene_sets=gene_sets,
                             format=file_format,
                             no_plot=~plot,
                             **kwargs)

        pre_res.res2d.sort_values(["NOM p-val"], inplace=True)
        pre_res.res2d.drop(["Name"], axis=1, inplace=True)

        if table:
            pre_res.res2d.to_csv(f"{file_name}.csv")

        if plot:
            res = pre_res.res2d.copy(deep=True)
            res = res[res['NOM p-val'] < p_value]
            for term in res.Term:
                name = term.split("(GO:")[1][:-1]
                gseaplot(rank_metric=pre_res.ranking,
                         term=term,
                         **pre_res.results[term],
                         ofname=f"{file_name}_GO_{name}.{file_format}")

        return pre_res.res2d



import sys
import warnings
import joblib
import pickle
import numpy as np
import pandas as pd
from sklearn.decomposition import LatentDirichletAllocation
import matplotlib.pyplot as plt
import seaborn as sns

sns.set_context('paper')
warnings.filterwarnings('ignore')

from Topyfic.topic import *


class TopModel:
    """
    A class that saved a model

    :param name: name of class
    :type name: str
    :param N: number of topics
    :type N: int
    :param gene_weights: dataframe that has weights of genes for each topics; genes are indexes and topics are columns
    :type gene_weights: pandas dataframe
    :param topics: dictionary contains all topics for the topmodel
    :type topics: Dictionary of Topics
    :param rlda: store reproducible LDA model
    :type rlda: sklearn.decomposition.LatentDirichletAllocation

    """

    def __init__(self,
                 name,
                 N,
                 gene_weights,
                 gene_information=None,
                 rlda=None):
        self.topics = dict()
        for i in range(N):
            topic_gene_weights = pd.DataFrame(gene_weights.iloc[:, i].values,
                                              index=gene_weights.index,
                                              columns=[f"Topic_{i + 1}"])
            self.topics[f"Topic_{i + 1}"] = Topic(topic_id=f"Topic_{i + 1}",
                                                  topic_gene_weights=topic_gene_weights,
                                                  gene_information=gene_information)
        self.name = name
        self.N = N
        self.rLDA = rlda

    def get_feature_name(self):
        """
        get feature(gene) name

        :return: list of feature(gene) name
        :rtype: list
        """
        topic1 = next(iter(self.topics.items()))[1]
        return topic1.gene_information.index.tolist()

    def save_rLDA_model(self, name='rLDA', save_path=""):
        """
        save Train class as a pickle file

        :param name: name of the pickle file (default: rLDA)
        :type name: str
        :param save_path: directory you want to use to save pickle file (default is saving near script)
        :type save_path: str
        """
        print(f"Saving rLDA model as {name}_{self.N}topics.joblib")

        joblib.dump(self.rLDA, f"{save_path}{name}_{self.N}topics.joblib", compress=3)

    def get_gene_weights(self):
        """
        get feature(gene) weights

        :return: dataframe contains feature(gene) weights; genes are indexes and topics are columns
        :rtype: pandas dataframe
        """
        gene_weights = pd.DataFrame(np.transpose(self.rLDA.components_),
                                    columns=[f'{self.name}_Topic_{i + 1}' for i in
                                             range(self.rLDA.components_.shape[0])],
                                    index=self.get_feature_name())

        return gene_weights

    def get_ranked_gene_weight(self):
        """
        get sorted feature(gene) weights. each value is gene and weights on each topics

        :return: dataframe contains feature(gene) and their weights; ranks are indexes and topics are columns
        :rtype: pandas dataframe
        """
        gene_weights = pd.DataFrame(np.transpose(self.rLDA.components_),
                                    columns=[f'{self.name}_Topic_{i + 1}' for i in
                                             range(self.rLDA.components_.shape[0])],
                                    index=self.get_feature_name())
        ranked_gene_weights = pd.DataFrame(columns=[f'{self.name}_Topic_{i + 1}' for i in
                                                    range(self.rLDA.components_.shape[0])],
                                           index=range(len(self.get_feature_name())))
        for col in gene_weights.columns:
            tmp = gene_weights.sort_values(by=col, ascending=False)
            tmp[col] = ";" + tmp[col].astype(str)
            tmp[col] = tmp.index.tolist() + tmp[col].astype(str)
            ranked_gene_weights[col] = tmp[col].values
        return ranked_gene_weights

    def get_top_model_attributes(self):
        """
        get top model attributes to be able to make sklearn.decomposition.LatentDirichletAllocation

        :return: three data frame which the first one is components, the second one is exp_dirichlet_component and
        the last one is combining the rest of LDA attributes which put them to gather as a dataframe
        :rtype: pandas dataframe, pandas dataframe, pandas dataframe
        """
        feature = self.get_feature_name()

        components = pd.DataFrame(self.rLDA.components_,
                                  index=[f"Topic_{i + 1}" for i in range(self.N)],
                                  columns=feature)

        exp_dirichlet_component = pd.DataFrame(self.rLDA.exp_dirichlet_component_,
                                               index=[f"Topic_{i + 1}" for i in range(self.N)],
                                               columns=feature)

        others = pd.DataFrame(
            index=[f"Topic_{i + 1}" for i in range(self.N)],
            columns=["n_batch_iter",
                     "n_features_in",
                     "n_iter",
                     "bound",
                     "doc_topic_prior",
                     "topic_word_prior"])

        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_batch_iter"] = self.rLDA.n_batch_iter_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_features_in"] = self.rLDA.n_features_in_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_iter"] = self.rLDA.n_iter_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "bound"] = self.rLDA.bound_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "doc_topic_prior"] = self.rLDA.doc_topic_prior_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "topic_word_prior"] = self.rLDA.topic_word_prior_

        return components, exp_dirichlet_component, others

    def gene_weight_rank_heatmap(self,
                                 genes=None,
                                 topics=None,
                                 show_rank=True,
                                 scale=None,
                                 save=True,
                                 show=True,
                                 figsize=None,
                                 file_format="pdf",
                                 file_name="gene_weight_rank_heatmap"):
        """
        plot selected genes weights and their ranks in selected topics

        :param genes: list of genes you want to see their weights (default: all genes)
        :type genes: list
        :param topics: list of topics
        :type topics: list
        :param show_rank: indicate if you want to show the rank of significant genes or not (default: True)
        :type show_rank: bool
        :param scale: indicate if you want to plot as log2, log10 or not (default: None which show actual value) other options is log2 and log10
        :scale scale: str
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_weight_rank_heatmap)
        :type file_name: str
        """
        if genes is None:
            genes = self.get_feature_name()
        elif not (set(genes) & set(self.get_feature_name())) == set(genes):
            sys.exit("some/all of genes are not part of topModel!")

        if topics is None:
            topics = [f'{self.name}_Topic_{i + 1}' for i in range(self.rLDA.components_.shape[0])]
        elif not (set(topics) & set(
                [f'{self.name}_Topic_{i + 1}' for i in range(self.rLDA.components_.shape[0])])) == set(topics):
            sys.exit("some/all of topics are not part of topModel!")

        if scale not in ["log2", "log10", None]:
            sys.exit("scale is not valid!")

        if figsize is None:
            figsize = (self.N * 1.5, len(genes))

        gene_weights = self.get_gene_weights()
        gene_weights = gene_weights.loc[:, topics]
        gene_rank = gene_weights.loc[genes, topics]
        for topic in topics:
            tmp = gene_weights.sort_values([topic], ascending=False)
            tmp['rank'] = range(1, tmp.shape[0] + 1)
            gene_rank.loc[genes, topic] = tmp.loc[genes, 'rank'].values

        gene_weights = self.get_gene_weights()
        gene_weights = gene_weights.loc[genes, topics]

        gene_weights = gene_weights.reindex(genes)
        gene_weights = gene_weights.reindex(topics, axis="columns")

        gene_rank = gene_rank.reindex(genes)
        gene_rank = gene_rank.reindex(topics, axis="columns")
        gene_rank = gene_rank.astype(int)
        gene_rank[gene_weights <= self.N] = 0
        gene_rank = gene_rank.astype(str)
        gene_rank[gene_rank == "0"] = ""

        fig, ax = plt.subplots(figsize=figsize,
                               facecolor='white')

        if scale == "log2":
            gene_weights = gene_weights.applymap(np.log2)
        if scale == "log10":
            gene_weights = gene_weights.applymap(np.log10)
        if scale is None:
            scale = "no"

        if show_rank:
            sns.heatmap(gene_weights,
                        cmap='viridis',
                        annot=gene_rank,
                        linewidths=.5,
                        fmt='',
                        ax=ax)
            ax.set_title(f'gene weights with rank ({scale} scale)')
        else:
            sns.heatmap(gene_weights,
                        cmap='viridis',
                        linewidths=.5,
                        ax=ax)
            ax.set_title('gene weights')

        if save:
            plt.savefig(f"{file_name}.{file_format}")
        if show:
            plt.show()
        else:
            plt.close()

    def save_topModel(self, name=None, save_path=""):
        """
        save TopModel class as a pickle file

        :param name: name of the pickle file (default: topModel_TopModel.name)
        :type name: str
        :param save_path: directory you want to use to save pickle file (default is saving near script)
        :type save_path: str
        """
        if name is None:
            name = f"topModel_{self.name}"
        print(f"Saving topModel as {name}.p")

        picklefile = open(f"{save_path}{name}.p", "wb")
        pickle.dump(self, picklefile)
        picklefile.close()



import click
from Topyfic.utils import *


@click.group()
def cli():
    pass


#### train model ####
@cli.command(name='train_model')
@click.option('--name',
              help='name of the Train class',
              required=True)
@click.option('--data',
              help='expression data embedded in anndata format along with cells and genes/region information',
              required=True)
@click.option('-k',
              help='number of topics to learn one LDA model using sklearn package',
              required=True)
@click.option('--n_runs',
              help='number of run to define rLDA model',
              default=100,
              show_default=True)
@click.option('--random_state_range',
              help=' list of random state, we used to run LDA models',
              default="range(n_run)",
              show_default=True)
@click.option('--n_thread',
              help='',
              default=5,
              show_default=True)
@click.option('--save_path',
              help='',
              default="",
              show_default=True)
def train_model_command(name,
                        data,
                        k,
                        n_runs=100,
                        random_state_range=None,
                        n_thread=5,
                        save_path=""):
    return train_model(name=name,
                       data=data,
                       k=k,
                       n_runs=n_runs,
                       random_state_range=random_state_range,
                       n_thread=n_thread,
                       save_path=save_path)


#### top model ####
@cli.command(name='make_topModel')
@click.option('--trains',
              help='list of train class',
              required=True)
@click.option('--data',
              help='expression data embedded in anndata format along with cells and genes/region information',
              required=True)
@click.option('--n_top_genes',
              help='Number of highly-variable genes to keep',
              default=50,
              show_default=True)
@click.option('--resolution',
              help='A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters.',
              default=1,
              show_default=True)
@click.option('--file_format',
              help='indicate the format of plot',
              default="pdf",
              show_default=True)
@click.option('--save_path',
              help='directory you want to use to save pickle file',
              default="",
              show_default=True)
def make_topModel_command(trains,
                          data,
                          n_top_genes=50,
                          resolution=1,
                          file_format="pdf",
                          save_path=""):
    return make_topModel(trains=trains,
                         data=data,
                         n_top_genes=n_top_genes,
                         resolution=resolution,
                         file_format=file_format,
                         save_path=save_path)


#### top model ####
@cli.command(name='make_topModel')
@click.option('--top_model',
              help='top model',
              required=True)
@click.option('--data',
              help='expression data embedded in anndata format along with cells and genes/region information',
              required=True)
@click.option('--colors_topics',
              help='dataframe that mapped colored to topics',
              default="assign color randomly",
              show_default=True)
@click.option('--save_path',
              help='directory you want to use to save pickle file',
              default="",
              show_default=True)
def make_analysis_class_command(top_model,
                                data,
                                colors_topics=None,
                                save_path=""):
    return make_analysis_class(top_model=top_model,
                               data=data,
                               colors_topics=colors_topics,
                               save_path=save_path)



import sys
import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import warnings
import joblib
import os
from multiprocessing import Pool
from itertools import repeat
import pickle
from sklearn.decomposition import LatentDirichletAllocation

from Topyfic.topic import *
from Topyfic.topModel import *

warnings.filterwarnings("ignore")


class Train:
    """
    A class used to train reproducible latent dirichlet allocation (rLDA) model

    :param name: name of the Train class
    :type name: str
    :param k: number of topics to learn one LDA model using sklearn package
    :type k: int
    :param n_runs: number of run to define rLDA model (default: 100)
    :type n_runs: int
    :param random_state_range: list of random state, we used to run LDA models (default: range(n_runs))
    :type random_state_range: list of int
    :param top_models: list of TopModel class to save all LDA models
    :type top_models: list of TopModel
    """

    def __init__(self,
                 name,
                 k,
                 n_runs=100,
                 random_state_range=None):

        if random_state_range is None:
            random_state_range = range(n_runs)
        elif n_runs != len(random_state_range):
            sys.exit("number of runs and length of random state did not match!")

        self.name = name
        self.k = k
        self.n_runs = n_runs
        self.random_state_range = random_state_range
        self.top_models = []

    def make_single_LDA_model(self, data, random_state, name, learning_method, batch_size, max_iter, n_jobs, kwargs):
        """
        train simple LDA model using sklearn package and embed it to TopModel class


        :param name: name of LDA model
        :type name: str
        :param data: processed expression data along with cells and genes/region information
        :type data: anndata
        :param random_state: Pass an int for reproducible results across multiple function calls
        :type random_state: int
        :param max_iter: The maximum number of passes over the training data (aka epochs) (default = 10)
        :type max_iter: int
        :param batch_size: Number of documents to use in each EM iteration. Only used in online learning. (default = 1000)
        :type batch_size: int
        :param learning_method: Method used to update _component. {‘batch’, ‘online’} (default=’online’)
        :type learning_method: str
        :param n_jobs: The number of jobs to use in the E-step. None means 1 unless in a `joblib.parallel_backend <https://joblib.readthedocs.io/en/latest/parallel.html#joblib.parallel_backend>`_ context. -1 means using all processors.  See `Glossary <https://scikit-learn.org/stable/glossary.html#term-n_jobs>`_ for more details. (default = None)
        :type n_jobs: int

        :return: LDA model embedded in TopModel class
        :rtype: TopModel
        """
        lda_model = LatentDirichletAllocation(n_components=self.k,
                                              random_state=random_state,
                                              learning_method=learning_method,
                                              batch_size=batch_size,
                                              max_iter=max_iter,
                                              n_jobs=n_jobs,
                                              **kwargs)

        lda_model.fit_transform(data.to_df().to_numpy())

        gene_weights = pd.DataFrame(np.transpose(lda_model.components_),
                                    columns=[f'Topic{i + 1}_R{random_state}' for i in range(self.k)],
                                    index=data.var.index.tolist())

        TopModel_lda_model = TopModel(name=f"{name}_{random_state}",
                                      N=gene_weights.shape[1],
                                      gene_weights=gene_weights,
                                      rlda=lda_model)

        return TopModel_lda_model

    def run_LDA_models(self, data, learning_method="online", batch_size=1000, max_iter=10, n_jobs=None, n_thread=1, **kwargs):
        """
        train LDA models

        :param max_iter: The maximum number of passes over the training data (aka epochs) (default = 10)
        :type max_iter: int
        :param batch_size: Number of documents to use in each EM iteration. Only used in online learning. (default = 1000)
        :type batch_size: int
        :param learning_method: Method used to update _component. {‘batch’, ‘online’} (default=’online’)
        :type learning_method: str
        :param data: expression data embedded in anndata format use to train LDA model
        :type data: anndata
        :param n_jobs: The number of jobs to use in the E-step. None means 1 unless in a `joblib.parallel_backend <https://joblib.readthedocs.io/en/latest/parallel.html#joblib.parallel_backend>`_ context. -1 means using all processors.  See `Glossary <https://scikit-learn.org/stable/glossary.html#term-n_jobs>`_ for more details. (default = None)
        :type n_jobs: int
        :param n_thread: number of threads you used to learn LDA models (default=1)
        :type n_thread: int
        :param **kwargs: other parameter in sklearn.decomposition.LatentDirichletAllocation function (more info: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html)
        :type **kwargs: dict

        :return: None
        :rtype: None
        """
        if n_thread is None:
            n_thread = self.n_runs

        self.top_models = Pool(processes=n_thread).starmap(self.make_single_LDA_model,
                                                           zip(repeat(data), self.random_state_range, repeat(self.name),
                                                               repeat(learning_method), repeat(batch_size),
                                                               repeat(max_iter), repeat(n_jobs), repeat(kwargs)))
        print(f"{self.n_runs} LDA models with {self.k} topics learned\n")

    def make_LDA_models_attributes(self):
        """
        make LDA attributes by combining all single LDA model attributes which you need to define LDA model (sklearn.decomposition.LatentDirichletAllocation)

        :return: three data frame which the first one is gathering all components from all LDA runs,
        the second one is exp_dirichlet_component from all LDA runs and
        the last one is combining the rest of LDA attributes which put them to gather as a dataframe
        :rtype: pandas dataframe, pandas dataframe, pandas dataframe
        """
        feature = self.top_models[0].get_feature_name()

        all_components = pd.DataFrame(
            index=[f"Topic{i + 1}_R{j}" for j in self.random_state_range for i in range(self.k)],
            columns=feature)

        all_exp_dirichlet_component = pd.DataFrame(
            index=[f"Topic{i + 1}_R{j}" for j in self.random_state_range for i in range(self.k)],
            columns=feature)

        all_others = pd.DataFrame(
            index=[f"Topic{i + 1}_R{j}" for j in self.random_state_range for i in range(self.k)],
            columns=["n_batch_iter",
                     "n_features_in",
                     "n_iter",
                     "bound",
                     "doc_topic_prior",
                     "topic_word_prior"])

        count = 0
        for random_state in self.random_state_range:
            components, exp_dirichlet_component, others = self.top_models[count].get_top_model_attributes()

            all_components.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], :] = components.values

            all_exp_dirichlet_component.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], :] = exp_dirichlet_component.values

            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "n_batch_iter"] = others.n_batch_iter.values
            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "n_features_in"] = others.n_features_in.values
            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "n_iter"] = others.n_iter.values
            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "bound"] = others.bound.values
            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "doc_topic_prior"] = others.doc_topic_prior.values
            all_others.loc[[f"Topic{i + 1}_R{random_state}" for i in range(self.k)], "topic_word_prior"] = others.topic_word_prior.values

            count = count + 1

        return all_components, all_exp_dirichlet_component, all_others

    def save_train(self, name=None, save_path=""):
        """
        save Train class as a pickle file

        :param name: name of the pickle file (default is train_Train.name)
        :type name: str
        :param save_path: directory you want to use to save pickle file (default is saving near script)
        :type save_path: str
        """
        if name is None:
            name = f"train_{self.name}"

        print(f"Saving train class as {name}.p")

        picklefile = open(f"{save_path}{name}.p", "wb")
        pickle.dump(self, picklefile)
        picklefile.close()



from setuptools import setup

setup(
    name='Topyfic',  # the name of your package
    packages=['Topyfic'],  # same as above
    version='v0.1.8',  # version number
    license='MIT',  # license type
    description='Topyfic is a Python package designed to identify reproducible latent dirichlet allocation (LDA) '
                'using leiden clustering and harmony for single cell epigenomics data',
    # short description
    author='Narges Rezaie',  # your name
    author_email='nargesrezaie80@gmail.com',  # your email
    url='https://github.com/mortazavilab/Topyfic',  # url to your git repo
    download_url='https://github.com/mortazavilab/Topyfic/archive/refs/tags/V0.1.8.tar.gz',  # link to the tar.gz file associated with this release
    keywords=['Cellular Programs', 'Latent Dirichlet allocation', 'single-cell multiome', 'single-cell RNA-seq',
              'gene regulatory network'],  #
    python_requires='>=3.8',
    install_requires=[  # these can also include >, <, == to enforce version compatibility
        'pandas',  # make sure the packages you put here are those NOT included in the
        'numpy',  # base python distribution
        'scipy',
        'scikit-learn>=0.24.2',
        'matplotlib',
        'seaborn',
        'joblib',
        'pytest',
        'scanpy',
        'anndata',
        'leidenalg',
        'networkx',
        'scanpy',
        'setuptools',
        'statsmodels',

    ],
    classifiers=[  # choose from here: https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research ',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
)



import pandas as pd
import numpy as np
import anndata
import sys
from scipy import stats
import warnings
import random
from statsmodels.stats.multitest import fdrcorrection
from scipy.cluster.hierarchy import ward, dendrogram, leaves_list

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import seaborn as sns

from Topyfic.topModel import *

sns.set_context('paper')
warnings.filterwarnings('ignore')


class Analysis:
    """
    A class used to investigate the topics and gene weights compositions

    :param Top_model: top model that used for analysing topics, gene weights compositions and calculate cell participation
    :type Top_model: TopModel
    :param colors_topics: dataframe that mapped colored to topics
    :type colors_topics: pandas dataframe
    :param cell_participation: anndata that stores cell participation along with cell information in obs
    :type cell_participation: anndata
    """

    def __init__(self,
                 Top_model,
                 colors_topics=None,
                 cell_participation=None):
        self.top_model = Top_model

        if colors_topics is None:
            colors = sns.color_palette("turbo", self.top_model.N).as_hex()

            def myfunction():
                return 0.1

            random.shuffle(colors, myfunction)
            index = list(self.top_model.topics.keys())
            self.colors_topics = pd.DataFrame({'colors': colors}, index=index)
        else:
            self.colors_topics = colors_topics

        self.cell_participation = cell_participation

    def calculate_cell_participation(self, data):
        """
        Calculate cell participation for give data

        :param data: processed expression data along with cells and genes/region information
        :type data: anndata
        """
        if self.cell_participation is not None:
            print("cell participation is not empty!")
            print("new cell participation will be replaced!")

        lda_output = self.top_model.rLDA.transform(data.X)
        cell_participation = pd.DataFrame(lda_output,
                                          columns=[f"Topic_{i + 1}" for i in range(self.top_model.N)],#list(self.top_model.topics.keys())
                                          index=data.obs.index)
        self.cell_participation = anndata.AnnData(cell_participation, obs=data.obs)

    def pie_structure_Chart(self,
                            level,
                            category=None,
                            ascending=None,
                            n=5,
                            save=True,
                            show=True,
                            figsize=None,
                            file_format="pdf",
                            file_name="piechart_topicAvgCell"):
        """
        plot pie charts that shows contribution of each topics to each category (i.e cell type)

        :param level: name of the column from cell_participation.obs
        :type level: str
        :param category: list of items you want to plot pie charts which are subsets of cell_participation.obs[level](default: all the unique items in cell_participation.obs[level])
        :type category: list of str
        :param ascending: for each pie chart on which order you want to sort your data (default is descending for all pie charts)
        :type ascending: list of bool
        :param n: number of topics you want to annotate in pie charts (default: 5)
        :type n: int
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: piechart_topicAvgCell)
        :type file_name: str
        """
        if self.cell_participation is None:
            sys.exit("self.cell_participation is empty, you need to fill it by using 'calculate_cell_participation'")

        if category is None:
            category = self.cell_participation.obs[level].unique().tolist()

        if figsize is None:
            figsize = (10 * (len(category) + 1), 10)

        fig, axs = plt.subplots(ncols=len(category) + 1,
                                figsize=figsize,
                                facecolor='white')

        colors = self.colors_topics

        if ascending is None:
            ascending = [False] * len(category)

        for i in range(len(category)):
            tissue = self.cell_participation.obs[self.cell_participation.obs[level] == category[i]]
            tmp = self.cell_participation.to_df().loc[tissue.index, :]
            order = tmp.mean().sort_values(ascending=False).index.tolist()
            index = tmp[order].sort_values(by=order, ascending=False).index.tolist()
            tmp = tmp.reindex(columns=order)
            tmp = tmp.reindex(index)
            colors = colors.reindex(order)
            labels = tmp.columns.tolist()
            labels[n:] = ['' for i in range(len(labels) - 5)]

            def make_autopct(values):
                def my_autopct(pct):
                    if pct > values[n] * 100:
                        return '{p:.0f}%'.format(p=pct)
                    else:
                        return ''

                return my_autopct

            axs[i].pie(tmp.mean(),
                       labels=labels,
                       colors=colors.colors.tolist(),
                       autopct=make_autopct(tmp.mean()),
                       wedgeprops={'linewidth': 0},
                       # labeldistance=0.8,
                       textprops={"fontsize": 25})

            axs[i].set_title(category[i], fontsize=40)

        colors = self.colors_topics
        handles = []
        for n in range(colors.shape[0]):
            patch = mpatches.Patch(color=colors.colors[n], label=colors.index[n])
            handles.append(patch)
            axs[len(category)].legend(loc='center left',
                                      title='Topic',
                                      ncol=3,
                                      handles=handles)
            axs[len(category)].axis('off')

        if save:
            fig.savefig(f"{file_name}.{file_format}")
        if show:
            plt.show()
        else:
            plt.close()

    def structure_plot(self,
                       level,
                       category,
                       topic_order=None,
                       ascending=None,
                       metaData=None,
                       metaData_palette=None,
                       width=None,
                       n=2,
                       order_cells=['hierarchy'],
                       save=True,
                       show=True,
                       figsize=None,
                       file_format="pdf",
                       file_name="structure_topicAvgCell"):
        """
        plot structure which shows contribution of each topics for each cells in given categories

        :param level: name of the column from cell_participation.obs
        :type level: str
        :param category: list of items you want to plot which are subsets of cell_participation.obs[level](default: all the unique items in cell_participation.obs[level])
        :type category: list of str
        :param topic_order: indicate if you want to have a specific order of topics which it should be name of topics. if None, it's gonna sort by cell participation
        :type topic_order: list of str
        :param ascending: for each structure plot on which order you want to sort your data (default is descending for all structure plot)
        :type ascending: list of bool
        :param metaData: if you want to add annotation for each cell add column name of that information (make sure you have that inforamtion in your cell_participation.obs)
        :type metaData: list
        :param metaData_palette: color palette for each metaData you add
        :type metaData_palette: dict
        :param width: width ratios of each category (default is based on the number of the cells we have in each category)
        :type width: list of int
        :param n: number of topics you want to annotate in pie charts (default: 5)
        :type n: int
        :param order_cells: determine which kind of sorting options you want to use ('sum', 'hierarchy', sort by metaData); sum: sort cells by sum of top n topics; hierarchy: sort data by doing hierarchical clustring; metaData sort by metaData (default: ['hierarchy'])
        :type order_cells: list
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: piechart_topicAvgCell)
        :type file_name: str
        """
        if figsize is None:
            figsize = (10 * (len(category) + 1), 10)

        check = ["hierarchy", "sum"] + metaData
        for order_cell in order_cells:
            if order_cell not in check:
                return "order cell was not valid"

        a = []
        for i in range(len(category)):
            a.append(self.cell_participation.obs[self.cell_participation.obs[level] == category[i]].shape[0])
        a.append(min(a) / 2)
        if width is None:
            width = a
        b = [8] + [1] * len(metaData)
        fig, axs = plt.subplots(nrows=len(metaData) + 1,
                                ncols=len(category) + 1,
                                figsize=figsize,
                                gridspec_kw={'width_ratios': width,
                                             'height_ratios': b},
                                facecolor='white')

        colors = self.colors_topics

        if ascending is None:
            ascending = [True] * len(category)

        for i in range(len(category)):
            tissue = self.cell_participation.obs[self.cell_participation.obs[level] == category[i]]
            tmp = self.cell_participation.to_df().loc[tissue.index, :]
            if topic_order is None:
                order = tmp.mean().sort_values(ascending=False).index.tolist()
            else:
                order = topic_order
            # tmp['non_major'] = 1 - tmp.sum(axis=1)
            # tmp.non_major[tmp['non_major'] < 0] = 0
            index = tmp[order].sort_values(by=order, ascending=False).index.tolist()
            tmp = tmp.reindex(index)
            if len(order_cells) == 1:
                order_cell = order_cells[0]
                if order_cell == 'hierarchy':
                    Z = ward(tmp)
                    index = leaves_list(Z).tolist()
                    tmp = tmp.iloc[index, :]
                elif order_cell == 'sum':
                    index = tmp[order[:n]].sum(axis=1).sort_values(ascending=ascending[i]).index.tolist()
                    tmp = tmp.reindex(index)
                elif order_cell in metaData:
                    index = tissue.sort_values(by=[order_cell], ascending=ascending[i]).index.tolist()
                    tmp = tmp.reindex(index)
            else:
                order_cell = order_cells[:-1]
                tissue.sort_values(by=order_cell, ascending=ascending[i], inplace=True)
                tmp = tmp.reindex(tissue.index)
                tissue['count'] = 1
                groups = tissue.groupby(order_cell).sum().reset_index()[order_cell + ['count']]
                groups.sort_values(by=order_cell, ascending=ascending[i], inplace=True)
                count = 0
                index = []
                for j in range(groups.shape[0]):
                    if groups.loc[j, 'count'] != 0:
                        sub = tissue.iloc[count:count + groups.loc[j, 'count'], :]
                        sub = tmp.loc[sub.index, :]
                        if sub.shape[0] > 1:
                            Z = ward(sub)
                            sub = leaves_list(Z).tolist()
                            sub = [x + count for x in sub]
                            index = index + sub
                        else:
                            index = index + [count]
                        count = count + groups.loc[j, 'count']

                tmp = tmp.iloc[index, :]

            # order.append('non_major')
            tmp = tmp.reindex(columns=order)

            colors = colors.reindex(order)
            if metaData is None:
                axs[i].stackplot(tmp.index.tolist(),
                                 tmp.T.to_numpy(),
                                 labels=tmp.columns.tolist(),
                                 colors=colors.colors.tolist(),
                                 linewidths=0)

                axs[i].xaxis.set_ticks([])
                axs[i].set_title(category[i], fontsize=40)
                axs[i].set_ylim(0, 1)
                axs[i].set_xlim(0, a[i])
                axs[0].set_ylabel("Topic proportion", fontsize=25)
            else:
                axs[0, i].stackplot(tmp.index.tolist(),
                                    tmp.T.to_numpy(),
                                    labels=tmp.columns.tolist(),
                                    colors=colors.colors.tolist(),
                                    linewidths=0)

                axs[0, i].xaxis.set_ticks([])
                axs[0, i].set_title(category[i], fontsize=40)
                axs[0, i].set_ylim(0, 1)
                axs[0, i].set_xlim(0, a[i])
                axs[0, 0].set_ylabel("Topic proportion", fontsize=25)

            tissue = tissue[metaData]
            tissue = tissue.reindex(tmp.index.tolist())
            for j in range(len(metaData)):
                if type(metaData_palette[metaData[j]]) == dict:
                    tissue.replace(metaData_palette[metaData[j]], inplace=True)

            x = [i for i in range(tmp.shape[0])]
            y = np.repeat(3000, len(x))
            for j in range(len(metaData)):
                color = tissue[metaData[j]].values
                if type(metaData_palette[metaData[j]]) == dict:
                    axs[j + 1, i].scatter(x, y, label=metaData_palette[metaData[j]],
                                          c=color, s=1000, marker="|", alpha=1,
                                          linewidths=1)
                else:
                    axs[j + 1, i].scatter(x, y, label=metaData_palette[metaData[j]],
                                          c=color, cmap=metaData_palette[metaData[j]].get_cmap(), s=1000, marker="|",
                                          alpha=1,
                                          linewidths=1)

                axs[j + 1, i].axis('off')
                axs[j + 1, i].set_xlim(0, a[i])

        colors = self.colors_topics

        handles = []
        for n in range(colors.shape[0]):
            patch = mpatches.Patch(color=colors.colors[n], label=colors.index[n])
            handles.append(patch)
            if metaData is None:
                axs[len(category)].legend(loc='center left',
                                          title='Topic',
                                          ncol=3,
                                          handles=handles)
                axs[len(category)].axis('off')
            else:
                axs[0, len(category)].legend(loc='center left',
                                             title='Topic',
                                             ncol=4,
                                             handles=handles)
                axs[0, len(category)].axis('off')

        for j in range(len(metaData)):
            handles = []
            if type(metaData_palette[metaData[j]]) == dict:
                for met in metaData_palette[metaData[j]].keys():
                    patch = mpatches.Patch(color=metaData_palette[metaData[j]][met], label=met)
                    handles.append(patch)
                    axs[j + 1, len(category)].legend(loc='center left',
                                                     title=metaData[j].capitalize(),
                                                     ncol=4,
                                                     handles=handles)
                    axs[j + 1, len(category)].axis('off')
            else:
                clb = fig.colorbar(mappable=metaData_palette[metaData[j]],
                                   ax=axs[j + 1, len(category)],
                                   orientation='horizontal',
                                   fraction=0.9)
                clb.ax.set_title(metaData[j].capitalize())
                axs[j + 1, len(category)].axis('off')

        if save:
            fig.savefig(f"{file_name}.{file_format}")
        if show:
            plt.show()
        else:
            plt.close()

    @staticmethod
    def convertDatTraits(data):
        """
        get data trait module base on samples information

        :return: a dataframe contains information in suitable format for plotting module trait relationship heatmap
        :rtype: pandas dataframe
        """
        datTraits = pd.DataFrame(index=data.index)
        for i in range(data.shape[1]):
            data.iloc[:, i] = data.iloc[:, i].astype(str)
            if len(np.unique(data.iloc[:, i])) == 2:
                datTraits[data.columns[i]] = data.iloc[:, i]
                org = np.unique(data.iloc[:, i]).tolist()
                rep = list(range(len(org)))
                datTraits.replace(to_replace=org, value=rep,
                                  inplace=True)
            elif len(np.unique(data.iloc[:, i])) > 2:
                for name in np.unique(data.iloc[:, i]):
                    datTraits[name] = data.iloc[:, i]
                    org = np.unique(data.iloc[:, i])
                    rep = np.repeat(0, len(org))
                    rep[np.where(org == name)] = 1
                    org = org.tolist()
                    rep = rep.tolist()
                    datTraits.replace(to_replace=org, value=rep, inplace=True)

        return datTraits

    def TopicTraitRelationshipHeatmap(self,
                                      metaData,
                                      save=True,
                                      show=True,
                                      file_format="pdf",
                                      file_name='topic-traitRelationships'):
        """
        plot topic-trait relationship heatmap

        :param metaData: traits you would like to see the relationship with topics (must be column name of cell_participation.obs)
        :type metaData: list
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: topic-traitRelationships)
        :type file_name: str
        """
        datTraits = Analysis.convertDatTraits(self.cell_participation.obs[metaData])

        topicsTraitCor = pd.DataFrame(index=self.cell_participation.to_df().columns,
                                      columns=datTraits.columns,
                                      dtype="float")
        topicsTraitPvalue = pd.DataFrame(index=self.cell_participation.to_df().columns,
                                         columns=datTraits.columns,
                                         dtype="float")

        for i in self.cell_participation.to_df().columns:
            for j in datTraits.columns:
                tmp = stats.pearsonr(self.cell_participation.to_df()[i], datTraits[j], alternative='greater')
                topicsTraitCor.loc[i, j] = tmp[0]
                topicsTraitPvalue.loc[i, j] = tmp[1]

        fig, ax = plt.subplots(figsize=(topicsTraitPvalue.shape[0] * 1.5,
                                        topicsTraitPvalue.shape[1] * 1.5), facecolor='white')

        xlabels = self.cell_participation.to_df().columns
        ylabels = datTraits.columns

        # print(topicsTraitPvalue)
        # for i in range(topicsTraitPvalue.shape[0]):
        #     rejected, tmp = fdrcorrection(topicsTraitPvalue.iloc[i, :])
        #     print(rejected)
        #     if not rejected:
        #         topicsTraitPvalue.iloc[i, :] = tmp
        # print(topicsTraitPvalue)

        # Loop over data dimensions and create text annotations.
        tmp_cor = topicsTraitCor.T.round(decimals=2)
        tmp_pvalue = topicsTraitPvalue.T.round(decimals=2)
        labels = (np.asarray(["{0}\n({1})".format(cor, pvalue)
                              for cor, pvalue in zip(tmp_cor.values.flatten(),
                                                     tmp_pvalue.values.flatten())])) \
            .reshape(topicsTraitCor.T.shape)

        sns.set(font_scale=1.5)
        res = sns.heatmap(topicsTraitCor.T, annot=labels, fmt="", cmap='RdBu_r',
                          vmin=-1, vmax=1, ax=ax, annot_kws={'size': 20, "weight": "bold"},
                          xticklabels=xlabels, yticklabels=ylabels)
        res.set_xticklabels(res.get_xmajorticklabels(), fontsize=20, fontweight="bold", rotation=90)
        res.set_yticklabels(res.get_ymajorticklabels(), fontsize=20, fontweight="bold")
        plt.yticks(rotation=0)
        ax.set_title(f"Topic-trait Relationships heatmap",
                     fontsize=30, fontweight="bold")
        ax.set_facecolor('white')

        if save:
            fig.savefig(f"{file_name}.{file_format}", bbox_inches='tight')
        if show:
            plt.show()
        else:
            plt.close()

    def extract_cells(self,
                      level,
                      category,
                      top_cells=0.05,
                      min_cell_participation=0.05,
                      min_cells=50,
                      file_name=None,
                      save=False):
        """
        extract subset of cells and cells participation with specific criteria

        :param level: name of the column from cell_participation.obs
        :type level: str
        :param category: list of items you want to plot which are subsets of cell_participation.obs[level](default: all the unique items in cell_participation.obs[level])
        :type category: list of str
        :param top_cells: fraction of the cells you want to be considers (default: 0.05)
        :type top_cells: float
        :param min_cell_participation: minimum cell participation each cells in each topics should have to be count (default: 0.05)
        :type min_cell_participation: float
        :param min_cells: minimum number of cells each topics should have to be reported (default: 50)
        :type min_cells: int
        :param file_name: name and path of the plot use for save (default: selectedCells_top{top_cells}_{min_cell_score}min_score_{min_cells}min_cells.csv and cellParticipation_selectedCells_top{top_cells}_{min_cell_score}min_score_{min_cells}min_cells.csv)
        :type file_name: str
        :param save: indicate if you want to save the data or not (default: False)
        :type save: bool

        :return: table contains cell ID that pass threshold for each topic, table contains cell particiaption for cells that pass threshold for each topic (same order as fist table)
        :rtype: pandas dataframe, pandas dataframe
        """
        cells = self.cell_participation.obs[self.cell_participation.obs[level].isin(category)]
        cell_topics = self.cell_participation.to_df().loc[cells.index, :]
        cell_topics = cell_topics[cell_topics > min_cell_participation]
        res = cell_topics.count() * top_cells
        res = res.round()

        selected_cell = pd.DataFrame(index=range(cells.shape[0]),
                                     columns=cell_topics.columns)
        selected_cell_participation = pd.DataFrame(index=range(cells.shape[0]),
                                                   columns=cell_topics.columns)

        for i in range(cell_topics.shape[1]):
            tmp = cell_topics.sort_values(cell_topics.columns[i], ascending=False)
            selected_cell[tmp.columns[i]][:int(res[tmp.columns[i]])] = tmp.index.tolist()[:int(res[tmp.columns[i]])]
            a = tmp[:int(res[tmp.columns[i]])]
            selected_cell_participation[tmp.columns[i]][:int(res[tmp.columns[i]])] = a[tmp.columns[i]].values.tolist()

        selected_cell.dropna(axis=0, how='all', inplace=True)
        selected_cell.dropna(axis=1, how='all', inplace=True)
        selected_cell = selected_cell[selected_cell.columns[selected_cell.count() > min_cells]]
        selected_cell.fillna("", inplace=True)

        selected_cell_participation.dropna(axis=0, how='all', inplace=True)
        selected_cell_participation.dropna(axis=1, how='all', inplace=True)
        selected_cell_participation = selected_cell_participation[
            selected_cell_participation.columns[selected_cell_participation.count() > min_cells]]
        selected_cell_participation.fillna("", inplace=True)

        if save:
            if file_name is None:
                selected_cell.to_csv(
                    f"selectedCells_top{top_cells}_{min_cell_participation}min_score_{min_cells}min_cells.csv")
                selected_cell_participation.to_csv(
                    f"cellParticipation_selectedCells_top{top_cells}_{min_cell_participation}min_score_{min_cells}min_cells.csv")
            else:
                selected_cell.to_csv(f"selectedCells_{file_name}.csv")
                selected_cell_participation.to_csv(f"cellParticipation_selectedCells_{file_name}.csv")

        else:
            return selected_cell, selected_cell_participation

    def average_cell_participation(self,
                                   label=None,
                                   color="blue",
                                   save=True,
                                   show=True,
                                   figsize=None,
                                   file_format="pdf",
                                   file_name="average_cell_participation"):
        """
        barplot showing average of cell participation in each topic

        :param label: fill with dictionary contain mapping new name for each topics to name you want to show if you want to change default topic name
        :type label: dict
        :param color: color of bar plot (default: blue)
        :type color: str
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: piechart_topicAvgCell)
        :type file_name: str
        """
        df = pd.DataFrame(self.cell_participation.to_df().mean())
        df.reset_index(inplace=True)
        if label is not None:
            df.replace(label, inplace=True)

        if figsize is None:
            figsize = (df.shape[0] / 1.5, 5)

        fig = plt.figure(figsize=figsize, facecolor='white')

        plt.bar(df['index'], df[0], color=color, width=0.5)

        plt.xlabel(self.top_model.name)
        plt.ylabel("Average cell participation")

        if save:
            fig.savefig(f"{file_name}.{file_format}", bbox_inches='tight')
        if show:
            plt.show()
        else:
            plt.close()

    def average_cell_participation_line_plot(self,
                                             topic,
                                             color,
                                             category,
                                             color_pallet=None,
                                             save=True,
                                             show=True,
                                             figsize=None,
                                             file_format="pdf",
                                             file_name="line_average_cell_participation"):
        """
        line plot showing average of cell participation in topic divided by two features of cells (i.e. cell type and time point)

        :param topic: name of the topic
        :type topic: str
        :param color: name of the feature you want to have one line per group of that (it should be column name of cell_participation.obs)
        :type color:str
        :type color_pallet: color of each category of color (if it None color assign randomly)
        :param color_pallet: dict
        :param category: name of the feature you want to have on x axis (it should be column name of cell_participation.obs)
        :type category: str
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: piechart_topicAvgCell)
        :type file_name: str
        """

        if topic not in self.top_model.topics.keys():
            sys.exit("topic is not valid!")

        if self.cell_participation is None:
            sys.exit("Cell participation is not calculated yet!")

        if category not in self.cell_participation.obs.columns:
            sys.exit("category is not valid!")

        if color not in self.cell_participation.obs.columns:
            sys.exit("color is not valid!")

        if color_pallet is None:
            color_pallet = dict()
            for name in self.cell_participation.obs[color].unique():
                color_pallet[name] = "#" + ''.join([random.choice('ABCDEF0123456789') for i in range(6)])

        res = pd.DataFrame(columns=self.cell_participation.obs[category].unique(),
                           index=self.cell_participation.obs[color].unique())

        for index in res.index:
            for col in res.columns:
                tmp = self.cell_participation[np.logical_and(self.cell_participation.obs[color] == index,
                                                             self.cell_participation.obs[category] == col)]
                res.loc[index, col] = tmp.to_df().mean()[topic]
        res.fillna(0, inplace=True)

        if figsize is None:
            figsize = (res.shape[0] / 1.5, 7)

        fig = plt.figure(figsize=figsize, facecolor='white')

        for index in res.index:
            plt.plot(res.columns.tolist(), res.loc[index, :].tolist(),
                     '-o',
                     label=index,
                     color=color_pallet[index])

        plt.legend(loc='center left', title=color, bbox_to_anchor=(1, 0.5))
        plt.xlabel(category)
        plt.ylabel("Average cell participation")
        plt.title(topic)

        if save:
            fig.savefig(f"{file_name}.{file_format}", bbox_inches='tight')
        if show:
            plt.show()
        else:
            plt.close()

    def plot_topic_composition(self,
                               category,
                               level="topic",
                               biotype="biotype",
                               label=False,
                               save=True,
                               show=True,
                               file_format="pdf",
                               file_name="gene_composition"):
        """
        plot gene composition dividing by gene biotype or topics

        :param category: topic name or gene biotype name you want to see gene composition for
        :type category: str
        :param level: indicate weather if you want to show it within each topic or gene biotype (options: "topic" or "gene_biotype") (default: topic)
        :type level: str
        :param biotype: name of the column in gene_weight to look for gene_biotype (default: biotype)
        :type biotype: str
        :param label: show label of each line within plot or not (default: False)
        :type label: bool
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_composition)
        :type file_name: str
        """
        if level not in ["topic", "gene_biotype"]:
            sys.exit("level is not correct! it should be topic or gene_biotype!")
        minWeight = np.min(self.top_model.get_gene_weights())
        gene_weights = self.top_model.get_gene_weights()
        gene_weights = gene_weights[gene_weights > minWeight]

        if level == "topic":
            gene_biotype = self.top_model.topics[category].gene_information[biotype]
            columns = gene_biotype.unique().tolist()
            df = gene_weights.sort_values([f"{self.top_model.name}_{category}"], axis=0, ascending=False)[
                f"{self.top_model.name}_{category}"]

            df = pd.DataFrame(df)
            df.dropna(axis=0, inplace=True)
            df.reset_index(inplace=True)
            df.columns = ['genes', 'weight']
            df.index = df['genes']
            df = pd.concat([df, gene_biotype], axis=1)
            df.dropna(axis=0, inplace=True)

            res = pd.DataFrame(columns=columns, dtype=int)
            for index, row in df.iterrows():
                # tmp.loc[0, 'weight'] = row['weight']
                if res.shape[0] == 0:
                    tmp = pd.DataFrame(0, index=[0], columns=columns, dtype=int)
                    tmp.loc[0, row[biotype]] = 1
                else:
                    tmp = pd.DataFrame(res.loc[res.shape[0] - 1]).T
                    tmp.reset_index(drop=True, inplace=True)
                    tmp.loc[0, row[biotype]] = res[row[biotype]][res.shape[0] - 1] + 1
                res = res.append(tmp, ignore_index=True)

        else:
            gene_weights.columns = [f'Topic_{i + 1}' for i in range(gene_weights.shape[1])]
            res = pd.DataFrame(columns=gene_weights.columns,
                               index=range(gene_weights.shape[0]),
                               dtype=int)
            gene_biotype = self.top_model.topics["Topic_1"].gene_information[biotype]
            columns = gene_biotype.unique().tolist()
            for col in gene_weights.columns:
                df = gene_weights.sort_values([col], axis=0, ascending=False)[col]

                df = pd.DataFrame(df)
                df.dropna(axis=0, inplace=True)
                df.reset_index(inplace=True)
                df.columns = ['genes', 'weight']
                df.index = df['genes']
                df = pd.concat([df, gene_biotype], axis=1)
                df.dropna(axis=0, inplace=True)

                sres = pd.DataFrame(columns=columns, dtype=int)
                for index, row in df.iterrows():
                    # tmp.loc[0, 'weight'] = row['weight']
                    if sres.shape[0] == 0:
                        tmp = pd.DataFrame(0, index=[0], columns=columns, dtype=int)
                        tmp.loc[0, row[biotype]] = 1
                    else:
                        tmp = pd.DataFrame(sres.loc[sres.shape[0] - 1]).T
                        tmp.reset_index(drop=True, inplace=True)
                        tmp.loc[0, row[biotype]] = sres[row[biotype]][sres.shape[0] - 1] + 1
                    sres = sres.append(tmp, ignore_index=True)

                res.loc[range(sres.shape[0]), col] = sres[category].values
            res.dropna(axis=0, how='all', inplace=True)

        sns.set_theme(style="white")
        fig, ax = plt.subplots(figsize=(max(res.shape[0] / 50, 10), max(res.shape[0] / 100, 10)),
                               facecolor='white')
        sns.set(font_scale=1)
        if level == "topic":
            sns.lineplot(data=res, dashes=False, ax=ax)
        else:
            sns.lineplot(data=res, dashes=False, ax=ax, palette=self.colors_topics.to_dict()['colors'])
        ax.set_yscale('log', base=2)
        if label:
            for line, name in zip(ax.lines, res.columns):
                y = int(line.get_ydata()[-1])
                if y == 0:
                    continue
                x = int(line.get_xdata()[-1])
                ax.text(x, y, name, color=line.get_color())
        ax.legend(loc='center left', bbox_to_anchor=(1.01, 0.5), ncol=1,
                  prop={'size': 15})
        ax.set_title(f"{self.top_model.name}_{category}", fontsize=25)
        ax.set_ylabel("#genes (log2)", fontsize=20)
        ax.set_xlabel("gene rank", fontsize=20)
        fig.tight_layout()

        if save:
            fig.savefig(f"{file_name}.{file_format}")
        if show:
            plt.show()
        else:
            plt.close()

    def save_analysis(self, name=None, save_path=""):
        """
        save Analysis class as a pickle file

        :param name: name of the pickle file (default: analysis_Analysis.top_model.name)
        :type name: str
        :param save_path: directory you want to use to save pickle file (default is saving near script)
        :type save_path: str
        """
        if name is None:
            name = f"analysis_{self.top_model.name}"
        print(f"Saving analysis class as {name}.p")

        picklefile = open(f"{save_path}{name}.p", "wb")
        pickle.dump(self, picklefile)
        picklefile.close()
